/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - sch-rnd (executable)
 *  Copyright (C) 2019..2022 Tibor 'Igor2' Palinkas
 *
 *  (Supported by NLnet NGI0 PET Fund in 2022)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

/* hidlib headers */
#include <librnd/core/unit.h>
#include <librnd/core/hid_init.h>
#include <librnd/core/hid.h>
#include <librnd/core/conf.h>
#include <librnd/core/hidlib.h>
#include <librnd/core/buildin.hidlib.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/plugins.h>

#include <librnd/core/actions.h>
#include <librnd/core/plugins.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/funchash.h>
#include <librnd/core/compat_lrealpath.h>

#include <libcschem/libcschem.h>
#include <libcschem/plug_io.h>
#include <libcschem/project.h>
#include <libcschem/abstract.h>
#include <libcschem/compile.h>
#include <libcschem/export.h>
#include <libcschem/message.h>
#include <libcschem/event.h>

#include "conf_core.h"
#include "crosshair.h"
#include "build_run.h"
#include "sheet.h"
#include "draw.h"
#include "font.h"
#include "emergency.h"

static const char *experimental = "Very early, pre-alpha version";
csch_project_t *prj; /* the project we are working on */

csch_chdr_t *csch_obj_clicked = NULL;

#define pup_buildins sch_rnd_buildins
#include "buildin.c"
#undef pup_buildins


TODO("multi: REMOVE")
rnd_hidlib_t *csch_hidlib;

static const char *EXPERIMENTAL = "The code is in alpha testing state yet.";

static const char *menu_file_paths[8];
static const char *menu_name_fmt = "menu-%s.lht";

#define CONF_USER_DIR "~/" DOT_SCH_RND


TODO("librnd: code dup with pcb-rnd; maybe we should have this in librnd?")
/* Figure out the canonical name of the executed program
   and fix up the defaults for various paths; returns exec prefix that
   should be saved in the config */
static char *main_path_init(char *argv0)
{
	size_t l;
	int haspath;
	char *t1, *t2;
	int found_bindir = 0, se = 0;
	char *exec_prefix = NULL;
	char *bindir = NULL, *tmp;


	/* see if argv0 has enough of a path to let lrealpath give the
	   real path.  This should be the case if you invoke pcb with
	   something like /usr/local/bin/pcb or ./pcb or ./foo/pcb
	   but if you just use pcb and it exists in your path, you'll
	   just get back pcb again. */
	haspath = (strchr(argv0, RND_DIR_SEPARATOR_C) != NULL);

#ifdef DEBUG
	printf("main_path_init (%s): haspath = %d\n", argv0, haspath);
#endif

	if (haspath) {
		bindir = rnd_lrealpath(argv0);
		if (bindir == NULL)
			bindir = rnd_strdup(argv0);
		found_bindir = 1;
	}
	else {
		char *path, *p, *tmps;
		struct stat sb;
		int r;

		tmps = getenv("PATH");

		if (tmps != NULL) {
			path = rnd_strdup(tmps);

			/* search through the font path for a font file */
			for (p = strtok(path, RND_PATH_DELIMETER); p && *p; p = strtok(NULL, RND_PATH_DELIMETER)) {
#ifdef DEBUG
				printf("Looking for %s in %s\n", argv0, p);
#endif
				if ((tmps = (char *) malloc((strlen(argv0) + strlen(p) + 2) * sizeof(char))) == NULL) {
					fprintf(stderr, "main_path_init():  malloc failed\n");
					exit(1);
				}
				sprintf(tmps, "%s%s%s", p, RND_DIR_SEPARATOR_S, argv0);
				r = stat(tmps, &sb);
				if (r == 0) {
#ifdef DEBUG
					printf("Found it:  \"%s\"\n", tmps);
#endif
					bindir = rnd_lrealpath(tmps);
					if (bindir == NULL)
						bindir = rnd_strdup(tmps);
					found_bindir = 1;
					free(tmps);
					break;
				}
				free(tmps);
			}
			free(path);
		}
	}

	if (found_bindir) {
		/* strip off the executable name leaving only the path */
		t2 = NULL;
		t1 = strchr(bindir, RND_DIR_SEPARATOR_C);
		while (t1 != NULL && *t1 != '\0') {
			t2 = t1;
			t1 = strchr(t2 + 1, RND_DIR_SEPARATOR_C);
		}
		if (t2 != NULL)
			*t2 = '\0';
	}
	else {
		/* we have failed to find out anything from argv[0] so fall back to the original install prefix */
		bindir = rnd_strdup(BINDIR);
	}

	/* now find the path to exec_prefix */
	l = strlen(bindir) + 1 + strlen(BINDIR_TO_EXECPREFIX) + 1;
	if ((exec_prefix = (char *) malloc(l * sizeof(char))) == NULL) {
		fprintf(stderr, "main_path_init():  malloc failed\n");
		exit(1);
	}
	sprintf(exec_prefix, "%s%s%s", bindir, RND_DIR_SEPARATOR_S, BINDIR_TO_EXECPREFIX);

	/* export the most important paths and data for child processes (e.g. parametric footprints) */
	tmp = rnd_concat(SCHSHAREDIR, "/footprint", NULL);
	se |= rnd_setenv("SCH_RND_VERSION",     SCH_VERSION,           1);
	se |= rnd_setenv("SCH_RND_REVISION",    SCH_REVISION,          1);
	se |= rnd_setenv("SCH_RND_PCBLIB",      tmp,                   1); /* for backward compatibility - do not use */
	se |= rnd_setenv("SCH_RND_FOOTPRINT",   tmp,                   1);
	se |= rnd_setenv("SCH_RND_SHARE",       SCHSHAREDIR,           1);
	se |= rnd_setenv("SCH_RND_LIB",         SCHLIBDIR,             1);
	se |= rnd_setenv("SCH_RND_EXEC_PREFIX", exec_prefix,           1);
	free(tmp);

	if (se != 0)
		fprintf(stderr, "WARNING: setenv() failed - external commands such as parametric footprints may not have a proper environment\n");

	menu_file_paths[0] = "./";
	menu_file_paths[1] = "~/.sch-rnd/";
	menu_file_paths[2] = rnd_concat(SCHCONFDIR, "/", NULL);
	menu_file_paths[3] = NULL;
	assert((sizeof(menu_file_paths)/sizeof(menu_file_paths[0])) > 3);

	rnd_pcbhl_conf_postproc(); /* get ~ (home dir) set */

TODO("fix conf")
#if 0
	rnd_app.conf_sysdir_path = SCHSHAREDIR;
	rnd_app.conf_sys_path = SCHSHAREDIR "/sch-rnd-conf.lht";
	rnd_app.lib_dir = SCHLIBDIR;
#else
/* temporary hack for running from ./ without internal version of the conf */
	rnd_app.conf_sysdir_path = "./";
	rnd_app.conf_sys_path = "./sch-rnd-conf.lht";
#endif

	rnd_app.conf_userdir_path = CONF_USER_DIR;
	rnd_app.conf_user_path = rnd_concat(CONF_USER_DIR, "/conf_core.lht", NULL);

	free(bindir);
	return exec_prefix;
}

static void main_path_uninit(void)
{
	/* const for all other parts of the code, but we had to concat (alloc) it above */
	free((char *)menu_file_paths[2]);
	free((char *)rnd_app.conf_user_path);
}


/* initialize signal and error handlers */
static void main_sighand_init(void)
{
#ifdef RND_HAVE_SIGHUP
	signal(SIGHUP, sch_rnd_catch_signal);
#endif
#ifdef RND_HAVE_SIGQUIT
	signal(SIGQUIT, sch_rnd_catch_signal);
#endif
#ifdef RND_HAVE_SIGTERM
	signal(SIGTERM, sch_rnd_catch_signal);
#endif
#ifdef RND_HAVE_SIGINT
	signal(SIGINT, sch_rnd_catch_signal);
#endif

#ifdef NDEBUG
/* so that we get a core dump on segfault in debug mode */
#	ifdef RND_HAVE_SIGABRT
		signal(SIGABRT, sch_rnd_catch_signal);
#	endif
#	ifdef RND_HAVE_SIGSEGV
		signal(SIGSEGV, sch_rnd_catch_signal);
#	endif
#endif

/* calling external program by popen() may cause a PIPE signal, so we ignore it */
#ifdef RND_HAVE_SIGPIPE
	signal(SIGPIPE, SIG_IGN);
#endif
}


static void gui_support_plugins(int load)
{
	static int loaded = 0;
	static pup_plugin_t *puphand;

	if (load && !loaded) {
		static const char *plugin_name = "sch_dialogs";
		int state = 0;
		loaded = 1;
		rnd_message(RND_MSG_DEBUG, "Loading GUI support plugin: '%s'\n", plugin_name);
		puphand = pup_load(&rnd_pup, (const char **)rnd_pup_paths, plugin_name, 0, &state);
		if (puphand == NULL)
			rnd_message(RND_MSG_ERROR, "Error: failed to load GUI support plugin '%s'\n-> expect missing widgets and dialog boxes\n", plugin_name);
	}
	if (!load && loaded && (puphand != NULL)) {
		pup_unload(&rnd_pup, puphand, NULL);
		loaded = 0;
		puphand = NULL;
	}
}

/* action table number of columns for a single action */
const char *sch_rnd_action_args[] = {
/*short, -long, action, help, hint-on-error */
	NULL, "-show-actions",    "PrintActions()",     "Print all available actions (human readable) and exit",   NULL,
	NULL, "-dump-actions",    "DumpActions()",      "Print all available actions (script readable) and exit",  NULL,
	NULL, "-dump-plugins",    "DumpPlugins()",      "Print all available plugins (script readable) and exit",  NULL,
	NULL, "-dump-plugindirs", "DumpPluginDirs()",   "Print directories plugins might be loaded from and exit", NULL,
	NULL, "-dump-oflags",     "DumpObjFlags()",     "Print object flags and exit",                             NULL,
	NULL, "-show-paths",      "PrintPaths()",       "Print all configured paths and exit",                     NULL,
	"V",  "-version",         "PrintVersion()",     "Print version info and exit",                             NULL,
	"V",  "-dump-version",    "DumpVersion()",      "Print version info in script readable format and exit",   NULL,
	NULL, "-copyright",       "PrintCopyright()",   "Print copyright and exit",                                NULL,
	NULL, NULL, NULL, NULL, NULL /* terminator */
};

#include "funchash_core.h"
#define action_entry(x) { #x, F_ ## x},
static rnd_funchash_table_t Functions[] = {
#include "funchash_core_list.h"
	{"F_END", F_END}
};


static void help(const char *topic)
{
	printf("No help yet\n");
}

static int do_export(int dump_abs, const char *exp_fmt)
{
	int res = 0;
	csch_abstract_t abs;

	csch_abstract_init(&abs);
	res |= csch_compile_project(prj, &abs, 0);

	if (dump_abs) {
		FILE *f = rnd_fopen(NULL, "ABS.txt", "w");
		csch_abstract_dump(&abs, f, "");
		fclose(f);
	}

	if (exp_fmt != NULL) {
		csch_project_export_sheets(prj, exp_fmt);

		{
			char epath[CSCH_PATH_MAX];
			csch_project_export_name(prj, epath, exp_fmt, NULL);
			csch_export_netlist(&abs, epath, exp_fmt);
		}
	}

	csch_abstract_uninit(&abs);
	return res;
}

static void gui_act_init(void)
{

}

extern void csch_main_act_init2(void);
extern void sch_rnd_file_act_init2(void);
extern void sch_rnd_draw_init2(void);
extern void sch_rnd_buffer_init2(void);
extern void sch_rnd_sheet_init2(void);


/*extern void rnd_main_act_init2(void);*/

static void sch_rnd_main_init(void)
{
	csch_init_actions();
	csch_main_act_init2();
	sch_rnd_sheet_init2();
	sch_rnd_file_act_init2();
	sch_rnd_font_init2();
	sch_rnd_draw_init2();
	sch_rnd_buffer_init2();
/*	rnd_main_act_init2();*/
}

extern void sch_rnd_draw_uninit(void);
extern void sch_rnd_buffer_uninit(void);

void sch_rnd_main_uninit(void)
{
	if (prj != NULL)
		csch_project_free(prj);
	rnd_funchash_uninit();
	csch_uninit();
	sch_rnd_draw_uninit();
	sch_rnd_font_uninit();
	sch_rnd_buffer_uninit();
	main_path_uninit();
	fgw_atexit();
}

int main(int argc, char *argv[])
{
	int n, res = 0, dump_abs = 0;
	const char *fmt = NULL, *exp_fmt = NULL;
	char *exec_prefix;
	rnd_main_args_t ga;
	csch_sheet_t *def_sheet = NULL;

	rnd_app.package = "sch-rnd";
	rnd_app.version = SCH_VERSION;
	rnd_app.url = "http://www.repo.hu/projects/cschem";

	rnd_app.menu_file_paths = menu_file_paths;
	rnd_app.menu_name_fmt   = menu_name_fmt;

	rnd_app.conf_core_postproc = sch_rnd_conf_core_postproc;
	rnd_app.conf_userdir_path = CONF_USER_DIR;
	rnd_app.conf_user_path = CONF_USER_DIR "/sch-rnd-conf.lht";
	rnd_app.dot_dir = DOT_SCH_RND;

	rnd_app.crosshair_move_to = csch_hidlib_crosshair_move_to;
	rnd_app.draw_attached = csch_draw_attached;
	rnd_app.expose_main = csch_expose_main;
	rnd_app.expose_preview = csch_expose_preview;

	rnd_fix_locale_and_env();
	exec_prefix = main_path_init(argv[0]);
	csch_init();

	rnd_main_args_init(&ga, argc, sch_rnd_action_args);
	rnd_hidlib_init1(conf_core_init);
	rnd_conf_set(RND_CFR_INTERNAL, "rc/path/exec_prefix", -1, exec_prefix, RND_POL_OVERWRITE);
	free(exec_prefix);

	csch_event_init_app();
	for(n = 1; n < argc; n++)
		n += rnd_main_args_add(&ga, argv[n], argv[n+1]);

	rnd_hidlib_init2(pup_buildins, sch_rnd_buildins);

	gui_act_init();
	sch_rnd_main_init();
	rnd_conf_set(RND_CFR_CLI, "editor/view/flip_y", 0, "1", RND_POL_OVERWRITE);

	rnd_funchash_set_table(Functions, RND_ENTRIES(Functions), NULL);

	/* Would be called from signal handlers too, using exit() */
	atexit(sch_rnd_emergency_save);


	rnd_hidlib_init3_auto();

	if (rnd_main_args_setup1(&ga) != 0) {
		sch_rnd_main_uninit();
		rnd_main_args_uninit(&ga);
		exit(1);
	}

/* Initialize actions only when the gui is already known so only the right
   one is registered (there can be only one GUI). */
TODO("action list");
/*#include "generated_lists.h"*/

	if (rnd_main_args_setup2(&ga, &res) != 0) {
		sch_rnd_main_uninit();
		rnd_main_args_uninit(&ga);
		exit(res);
	}

	for(n = 0; ga.hid_argc > 0; n++, ga.hid_argc--) {
		const char *fn = ga.hid_argv[n];
		if (prj == NULL)
			prj = csch_load_project_by_sheet_name(fn);

		if (csch_project_load_sheet(prj, fn, fmt, &def_sheet) != 0) {
			if (rnd_main_exporting) {
				csch_message(CSCH_MSG_ERROR, "Can not load file '%s' (specified on command line) for exporting or printing\n", fn);
				rnd_log_print_uninit_errs("Export load error");
				exit(1);
			}
		}
		if (def_sheet != NULL)
			sch_rnd_sheet_postproc(def_sheet);
	}

	if (def_sheet == NULL) {
		if (prj == NULL)
			prj = csch_project_alloc();
		def_sheet = sch_rnd_sheet_new(prj);
		if (def_sheet == NULL) {
			fprintf(stderr, "Failed to create new sheet (by loading the default sheet)\n");
			sch_rnd_main_uninit();
			rnd_main_args_uninit(&ga);
			exit(1);
		}
	}

	main_sighand_init();

	if (rnd_main_exported(&ga, &def_sheet->hidlib, 0)) {
		sch_rnd_main_uninit();
		rnd_main_args_uninit(&ga);
		exit(0);
	}

#if 0
	if (export_only)
		res = do_export(dump_abs, exp_fmt);
#endif

	csch_crosshair_gui_init();


	rnd_event(&def_sheet->hidlib, RND_EVENT_BOARD_CHANGED, "i", 0);

	sch_rnd_enable_autosave();

	/* main loop */
	if (RND_HAVE_GUI_ATTR_DLG)
		gui_support_plugins(1);
	if (EXPERIMENTAL != NULL) {
		csch_message(CSCH_MSG_ERROR, "******************************** IMPORTANT ********************************\n");
		csch_message(CSCH_MSG_ERROR, "This revision of sch-rnd is experimental, unstable, do NOT attempt to use\n");
		csch_message(CSCH_MSG_ERROR, "it for production. The reason for this state is:\n");
		csch_message(CSCH_MSG_ERROR, "%s\n", EXPERIMENTAL);
		csch_message(CSCH_MSG_ERROR, "******************************** IMPORTANT ********************************\n");
	}

	csch_hidlib = &def_sheet->hidlib;
	rnd_mainloop_interactive(&ga, &def_sheet->hidlib);

	csch_crosshair_gui_uninit();

	sch_rnd_main_uninit();
	rnd_main_args_uninit(&ga);
	return res;
}
